#include "sym_kernels.h"
#include "device.h"
#include "debug.h"

#if CUDART_VERSION >= 0
  __global__ void ncclSymkDevKernel_AllGather_LL(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllGather_LL = nullptr;
#endif
#if CUDART_VERSION >= 0
  __global__ void ncclSymkDevKernel_AllGather_LL_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllGather_LL_profile = nullptr;
#endif
#if CUDART_VERSION >= 0
  __global__ void ncclSymkDevKernel_AllGather_LLMC(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllGather_LLMC = nullptr;
#endif
#if CUDART_VERSION >= 0
  __global__ void ncclSymkDevKernel_AllGather_LLMC_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllGather_LLMC_profile = nullptr;
#endif
#if CUDART_VERSION >= 0
  __global__ void ncclSymkDevKernel_AllGather_ST(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllGather_ST = nullptr;
#endif
#if CUDART_VERSION >= 0
  __global__ void ncclSymkDevKernel_AllGather_ST_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllGather_ST_profile = nullptr;
#endif
#if CUDART_VERSION >= 0
  __global__ void ncclSymkDevKernel_AllGather_STMC(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllGather_STMC = nullptr;
#endif
#if CUDART_VERSION >= 0
  __global__ void ncclSymkDevKernel_AllGather_STMC_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllGather_STMC_profile = nullptr;
#endif
#if CUDART_VERSION >= 12000
  __global__ void ncclSymkDevKernel_AllGather_TmaST(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllGather_TmaST = nullptr;
#endif
#if CUDART_VERSION >= 12000
  __global__ void ncclSymkDevKernel_AllGather_TmaST_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllGather_TmaST_profile = nullptr;
#endif
#if CUDART_VERSION >= 12000
  __global__ void ncclSymkDevKernel_AllGather_TmaSTMC(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllGather_TmaSTMC = nullptr;
#endif
#if CUDART_VERSION >= 12000
  __global__ void ncclSymkDevKernel_AllGather_TmaSTMC_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllGather_TmaSTMC_profile = nullptr;
#endif
#if CUDART_VERSION >= 0
  __global__ void ncclSymkDevKernel_AllGather_RailRing_LsaSTMC(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllGather_RailRing_LsaSTMC = nullptr;
#endif
#if CUDART_VERSION >= 0
  __global__ void ncclSymkDevKernel_AllGather_RailRing_LsaSTMC_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllGather_RailRing_LsaSTMC_profile = nullptr;
#endif
#if CUDART_VERSION >= 0
  __global__ void ncclSymkDevKernel_AllReduce_AGxLL_R_sum_f32(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllReduce_AGxLL_R_sum_f32 = nullptr;
#endif
#if CUDART_VERSION >= 0
  __global__ void ncclSymkDevKernel_AllReduce_AGxLL_R_sum_f32_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllReduce_AGxLL_R_sum_f32_profile = nullptr;
#endif
#if CUDART_VERSION >= 12010
  __global__ void ncclSymkDevKernel_AllReduce_AGxLLMC_R_sum_f32(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllReduce_AGxLLMC_R_sum_f32 = nullptr;
#endif
#if CUDART_VERSION >= 12010
  __global__ void ncclSymkDevKernel_AllReduce_AGxLLMC_R_sum_f32_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllReduce_AGxLLMC_R_sum_f32_profile = nullptr;
#endif
#if CUDART_VERSION >= 0
  __global__ void ncclSymkDevKernel_AllReduce_RSxLD_AGxST_sum_f32(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllReduce_RSxLD_AGxST_sum_f32 = nullptr;
#endif
#if CUDART_VERSION >= 0
  __global__ void ncclSymkDevKernel_AllReduce_RSxLD_AGxST_sum_f32_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllReduce_RSxLD_AGxST_sum_f32_profile = nullptr;
#endif
#if CUDART_VERSION >= 12010
  __global__ void ncclSymkDevKernel_AllReduce_RSxLDMC_AGxSTMC_sum_f32(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllReduce_RSxLDMC_AGxSTMC_sum_f32 = nullptr;
#endif
#if CUDART_VERSION >= 12010
  __global__ void ncclSymkDevKernel_AllReduce_RSxLDMC_AGxSTMC_sum_f32_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllReduce_RSxLDMC_AGxSTMC_sum_f32_profile = nullptr;
#endif
#if CUDART_VERSION >= 12000
  __global__ void ncclSymkDevKernel_AllReduce_RSxTmaLD_AGxTmaST_sum_f32(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllReduce_RSxTmaLD_AGxTmaST_sum_f32 = nullptr;
#endif
#if CUDART_VERSION >= 12000
  __global__ void ncclSymkDevKernel_AllReduce_RSxTmaLD_AGxTmaST_sum_f32_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllReduce_RSxTmaLD_AGxTmaST_sum_f32_profile = nullptr;
#endif
#if CUDART_VERSION >= 0
  __global__ void ncclSymkDevKernel_ReduceScatter_LL_sum_f32(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LL_sum_f32 = nullptr;
#endif
#if CUDART_VERSION >= 0
  __global__ void ncclSymkDevKernel_ReduceScatter_LL_sum_f32_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LL_sum_f32_profile = nullptr;
#endif
#if CUDART_VERSION >= 0
  __global__ void ncclSymkDevKernel_ReduceScatter_LD_sum_f32(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LD_sum_f32 = nullptr;
#endif
#if CUDART_VERSION >= 0
  __global__ void ncclSymkDevKernel_ReduceScatter_LD_sum_f32_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LD_sum_f32_profile = nullptr;
#endif
#if CUDART_VERSION >= 12010
  __global__ void ncclSymkDevKernel_ReduceScatter_LDMC_sum_f32(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LDMC_sum_f32 = nullptr;
#endif
#if CUDART_VERSION >= 12010
  __global__ void ncclSymkDevKernel_ReduceScatter_LDMC_sum_f32_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LDMC_sum_f32_profile = nullptr;
#endif
#if CUDART_VERSION >= 12000
  __global__ void ncclSymkDevKernel_ReduceScatter_TmaLD_sum_f32(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_TmaLD_sum_f32 = nullptr;
#endif
#if CUDART_VERSION >= 12000
  __global__ void ncclSymkDevKernel_ReduceScatter_TmaLD_sum_f32_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_TmaLD_sum_f32_profile = nullptr;
#endif
#if CUDART_VERSION >= 0
  __global__ void ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_sum_f32(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_sum_f32 = nullptr;
#endif
#if CUDART_VERSION >= 0
  __global__ void ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_sum_f32_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_sum_f32_profile = nullptr;
#endif
#if CUDART_VERSION >= 12010
  __global__ void ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_sum_f32(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_sum_f32 = nullptr;
#endif
#if CUDART_VERSION >= 12010
  __global__ void ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_sum_f32_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_sum_f32_profile = nullptr;
#endif
#if CUDART_VERSION >= 0
  __global__ void ncclSymkDevKernel_AllReduce_AGxLL_R_sum_f16(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllReduce_AGxLL_R_sum_f16 = nullptr;
#endif
#if CUDART_VERSION >= 0
  __global__ void ncclSymkDevKernel_AllReduce_AGxLL_R_sum_f16_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllReduce_AGxLL_R_sum_f16_profile = nullptr;
#endif
#if CUDART_VERSION >= 12010
  __global__ void ncclSymkDevKernel_AllReduce_AGxLLMC_R_sum_f16(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllReduce_AGxLLMC_R_sum_f16 = nullptr;
#endif
#if CUDART_VERSION >= 12010
  __global__ void ncclSymkDevKernel_AllReduce_AGxLLMC_R_sum_f16_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllReduce_AGxLLMC_R_sum_f16_profile = nullptr;
#endif
#if CUDART_VERSION >= 0
  __global__ void ncclSymkDevKernel_AllReduce_RSxLD_AGxST_sum_f16(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllReduce_RSxLD_AGxST_sum_f16 = nullptr;
#endif
#if CUDART_VERSION >= 0
  __global__ void ncclSymkDevKernel_AllReduce_RSxLD_AGxST_sum_f16_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllReduce_RSxLD_AGxST_sum_f16_profile = nullptr;
#endif
#if CUDART_VERSION >= 12010
  __global__ void ncclSymkDevKernel_AllReduce_RSxLDMC_AGxSTMC_sum_f16(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllReduce_RSxLDMC_AGxSTMC_sum_f16 = nullptr;
#endif
#if CUDART_VERSION >= 12010
  __global__ void ncclSymkDevKernel_AllReduce_RSxLDMC_AGxSTMC_sum_f16_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllReduce_RSxLDMC_AGxSTMC_sum_f16_profile = nullptr;
#endif
#if CUDART_VERSION >= 12000
  __global__ void ncclSymkDevKernel_AllReduce_RSxTmaLD_AGxTmaST_sum_f16(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllReduce_RSxTmaLD_AGxTmaST_sum_f16 = nullptr;
#endif
#if CUDART_VERSION >= 12000
  __global__ void ncclSymkDevKernel_AllReduce_RSxTmaLD_AGxTmaST_sum_f16_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllReduce_RSxTmaLD_AGxTmaST_sum_f16_profile = nullptr;
#endif
#if CUDART_VERSION >= 0
  __global__ void ncclSymkDevKernel_ReduceScatter_LL_sum_f16(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LL_sum_f16 = nullptr;
#endif
#if CUDART_VERSION >= 0
  __global__ void ncclSymkDevKernel_ReduceScatter_LL_sum_f16_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LL_sum_f16_profile = nullptr;
#endif
#if CUDART_VERSION >= 0
  __global__ void ncclSymkDevKernel_ReduceScatter_LD_sum_f16(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LD_sum_f16 = nullptr;
#endif
#if CUDART_VERSION >= 0
  __global__ void ncclSymkDevKernel_ReduceScatter_LD_sum_f16_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LD_sum_f16_profile = nullptr;
#endif
#if CUDART_VERSION >= 12010
  __global__ void ncclSymkDevKernel_ReduceScatter_LDMC_sum_f16(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LDMC_sum_f16 = nullptr;
#endif
#if CUDART_VERSION >= 12010
  __global__ void ncclSymkDevKernel_ReduceScatter_LDMC_sum_f16_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LDMC_sum_f16_profile = nullptr;
#endif
#if CUDART_VERSION >= 12000
  __global__ void ncclSymkDevKernel_ReduceScatter_TmaLD_sum_f16(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_TmaLD_sum_f16 = nullptr;
#endif
#if CUDART_VERSION >= 12000
  __global__ void ncclSymkDevKernel_ReduceScatter_TmaLD_sum_f16_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_TmaLD_sum_f16_profile = nullptr;
#endif
#if CUDART_VERSION >= 0
  __global__ void ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_sum_f16(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_sum_f16 = nullptr;
#endif
#if CUDART_VERSION >= 0
  __global__ void ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_sum_f16_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_sum_f16_profile = nullptr;
#endif
#if CUDART_VERSION >= 12010
  __global__ void ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_sum_f16(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_sum_f16 = nullptr;
#endif
#if CUDART_VERSION >= 12010
  __global__ void ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_sum_f16_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_sum_f16_profile = nullptr;
#endif
#if CUDART_VERSION >= 11000
  __global__ void ncclSymkDevKernel_AllReduce_AGxLL_R_sum_bf16(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllReduce_AGxLL_R_sum_bf16 = nullptr;
#endif
#if CUDART_VERSION >= 11000
  __global__ void ncclSymkDevKernel_AllReduce_AGxLL_R_sum_bf16_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllReduce_AGxLL_R_sum_bf16_profile = nullptr;
#endif
#if CUDART_VERSION >= 12010
  __global__ void ncclSymkDevKernel_AllReduce_AGxLLMC_R_sum_bf16(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllReduce_AGxLLMC_R_sum_bf16 = nullptr;
#endif
#if CUDART_VERSION >= 12010
  __global__ void ncclSymkDevKernel_AllReduce_AGxLLMC_R_sum_bf16_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllReduce_AGxLLMC_R_sum_bf16_profile = nullptr;
#endif
#if CUDART_VERSION >= 11000
  __global__ void ncclSymkDevKernel_AllReduce_RSxLD_AGxST_sum_bf16(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllReduce_RSxLD_AGxST_sum_bf16 = nullptr;
#endif
#if CUDART_VERSION >= 11000
  __global__ void ncclSymkDevKernel_AllReduce_RSxLD_AGxST_sum_bf16_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllReduce_RSxLD_AGxST_sum_bf16_profile = nullptr;
#endif
#if CUDART_VERSION >= 12010
  __global__ void ncclSymkDevKernel_AllReduce_RSxLDMC_AGxSTMC_sum_bf16(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllReduce_RSxLDMC_AGxSTMC_sum_bf16 = nullptr;
#endif
#if CUDART_VERSION >= 12010
  __global__ void ncclSymkDevKernel_AllReduce_RSxLDMC_AGxSTMC_sum_bf16_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllReduce_RSxLDMC_AGxSTMC_sum_bf16_profile = nullptr;
#endif
#if CUDART_VERSION >= 12000
  __global__ void ncclSymkDevKernel_AllReduce_RSxTmaLD_AGxTmaST_sum_bf16(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllReduce_RSxTmaLD_AGxTmaST_sum_bf16 = nullptr;
#endif
#if CUDART_VERSION >= 12000
  __global__ void ncclSymkDevKernel_AllReduce_RSxTmaLD_AGxTmaST_sum_bf16_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllReduce_RSxTmaLD_AGxTmaST_sum_bf16_profile = nullptr;
#endif
#if CUDART_VERSION >= 11000
  __global__ void ncclSymkDevKernel_ReduceScatter_LL_sum_bf16(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LL_sum_bf16 = nullptr;
#endif
#if CUDART_VERSION >= 11000
  __global__ void ncclSymkDevKernel_ReduceScatter_LL_sum_bf16_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LL_sum_bf16_profile = nullptr;
#endif
#if CUDART_VERSION >= 11000
  __global__ void ncclSymkDevKernel_ReduceScatter_LD_sum_bf16(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LD_sum_bf16 = nullptr;
#endif
#if CUDART_VERSION >= 11000
  __global__ void ncclSymkDevKernel_ReduceScatter_LD_sum_bf16_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LD_sum_bf16_profile = nullptr;
#endif
#if CUDART_VERSION >= 12010
  __global__ void ncclSymkDevKernel_ReduceScatter_LDMC_sum_bf16(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LDMC_sum_bf16 = nullptr;
#endif
#if CUDART_VERSION >= 12010
  __global__ void ncclSymkDevKernel_ReduceScatter_LDMC_sum_bf16_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LDMC_sum_bf16_profile = nullptr;
#endif
#if CUDART_VERSION >= 12000
  __global__ void ncclSymkDevKernel_ReduceScatter_TmaLD_sum_bf16(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_TmaLD_sum_bf16 = nullptr;
#endif
#if CUDART_VERSION >= 12000
  __global__ void ncclSymkDevKernel_ReduceScatter_TmaLD_sum_bf16_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_TmaLD_sum_bf16_profile = nullptr;
#endif
#if CUDART_VERSION >= 11000
  __global__ void ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_sum_bf16(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_sum_bf16 = nullptr;
#endif
#if CUDART_VERSION >= 11000
  __global__ void ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_sum_bf16_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_sum_bf16_profile = nullptr;
#endif
#if CUDART_VERSION >= 12010
  __global__ void ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_sum_bf16(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_sum_bf16 = nullptr;
#endif
#if CUDART_VERSION >= 12010
  __global__ void ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_sum_bf16_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_sum_bf16_profile = nullptr;
#endif
#if CUDART_VERSION >= 11080
  __global__ void ncclSymkDevKernel_AllReduce_AGxLL_R_sum_f8e4m3(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllReduce_AGxLL_R_sum_f8e4m3 = nullptr;
#endif
#if CUDART_VERSION >= 11080
  __global__ void ncclSymkDevKernel_AllReduce_AGxLL_R_sum_f8e4m3_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllReduce_AGxLL_R_sum_f8e4m3_profile = nullptr;
#endif
#if CUDART_VERSION >= 12010
  __global__ void ncclSymkDevKernel_AllReduce_AGxLLMC_R_sum_f8e4m3(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllReduce_AGxLLMC_R_sum_f8e4m3 = nullptr;
#endif
#if CUDART_VERSION >= 12010
  __global__ void ncclSymkDevKernel_AllReduce_AGxLLMC_R_sum_f8e4m3_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllReduce_AGxLLMC_R_sum_f8e4m3_profile = nullptr;
#endif
#if CUDART_VERSION >= 11080
  __global__ void ncclSymkDevKernel_AllReduce_RSxLD_AGxST_sum_f8e4m3(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllReduce_RSxLD_AGxST_sum_f8e4m3 = nullptr;
#endif
#if CUDART_VERSION >= 11080
  __global__ void ncclSymkDevKernel_AllReduce_RSxLD_AGxST_sum_f8e4m3_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllReduce_RSxLD_AGxST_sum_f8e4m3_profile = nullptr;
#endif
#if CUDART_VERSION >= 12070
  __global__ void ncclSymkDevKernel_AllReduce_RSxLDMC_AGxSTMC_sum_f8e4m3(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllReduce_RSxLDMC_AGxSTMC_sum_f8e4m3 = nullptr;
#endif
#if CUDART_VERSION >= 12070
  __global__ void ncclSymkDevKernel_AllReduce_RSxLDMC_AGxSTMC_sum_f8e4m3_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllReduce_RSxLDMC_AGxSTMC_sum_f8e4m3_profile = nullptr;
#endif
#if CUDART_VERSION >= 12000
  __global__ void ncclSymkDevKernel_AllReduce_RSxTmaLD_AGxTmaST_sum_f8e4m3(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllReduce_RSxTmaLD_AGxTmaST_sum_f8e4m3 = nullptr;
#endif
#if CUDART_VERSION >= 12000
  __global__ void ncclSymkDevKernel_AllReduce_RSxTmaLD_AGxTmaST_sum_f8e4m3_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllReduce_RSxTmaLD_AGxTmaST_sum_f8e4m3_profile = nullptr;
#endif
#if CUDART_VERSION >= 11080
  __global__ void ncclSymkDevKernel_ReduceScatter_LL_sum_f8e4m3(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LL_sum_f8e4m3 = nullptr;
#endif
#if CUDART_VERSION >= 11080
  __global__ void ncclSymkDevKernel_ReduceScatter_LL_sum_f8e4m3_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LL_sum_f8e4m3_profile = nullptr;
#endif
#if CUDART_VERSION >= 11080
  __global__ void ncclSymkDevKernel_ReduceScatter_LD_sum_f8e4m3(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LD_sum_f8e4m3 = nullptr;
#endif
#if CUDART_VERSION >= 11080
  __global__ void ncclSymkDevKernel_ReduceScatter_LD_sum_f8e4m3_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LD_sum_f8e4m3_profile = nullptr;
#endif
#if CUDART_VERSION >= 12070
  __global__ void ncclSymkDevKernel_ReduceScatter_LDMC_sum_f8e4m3(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LDMC_sum_f8e4m3 = nullptr;
#endif
#if CUDART_VERSION >= 12070
  __global__ void ncclSymkDevKernel_ReduceScatter_LDMC_sum_f8e4m3_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LDMC_sum_f8e4m3_profile = nullptr;
#endif
#if CUDART_VERSION >= 12000
  __global__ void ncclSymkDevKernel_ReduceScatter_TmaLD_sum_f8e4m3(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_TmaLD_sum_f8e4m3 = nullptr;
#endif
#if CUDART_VERSION >= 12000
  __global__ void ncclSymkDevKernel_ReduceScatter_TmaLD_sum_f8e4m3_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_TmaLD_sum_f8e4m3_profile = nullptr;
#endif
#if CUDART_VERSION >= 11080
  __global__ void ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_sum_f8e4m3(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_sum_f8e4m3 = nullptr;
#endif
#if CUDART_VERSION >= 11080
  __global__ void ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_sum_f8e4m3_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_sum_f8e4m3_profile = nullptr;
#endif
#if CUDART_VERSION >= 12070
  __global__ void ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_sum_f8e4m3(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_sum_f8e4m3 = nullptr;
#endif
#if CUDART_VERSION >= 12070
  __global__ void ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_sum_f8e4m3_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_sum_f8e4m3_profile = nullptr;
#endif
#if CUDART_VERSION >= 11080
  __global__ void ncclSymkDevKernel_AllReduce_AGxLL_R_sum_f8e5m2(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllReduce_AGxLL_R_sum_f8e5m2 = nullptr;
#endif
#if CUDART_VERSION >= 11080
  __global__ void ncclSymkDevKernel_AllReduce_AGxLL_R_sum_f8e5m2_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllReduce_AGxLL_R_sum_f8e5m2_profile = nullptr;
#endif
#if CUDART_VERSION >= 12010
  __global__ void ncclSymkDevKernel_AllReduce_AGxLLMC_R_sum_f8e5m2(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllReduce_AGxLLMC_R_sum_f8e5m2 = nullptr;
#endif
#if CUDART_VERSION >= 12010
  __global__ void ncclSymkDevKernel_AllReduce_AGxLLMC_R_sum_f8e5m2_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllReduce_AGxLLMC_R_sum_f8e5m2_profile = nullptr;
#endif
#if CUDART_VERSION >= 11080
  __global__ void ncclSymkDevKernel_AllReduce_RSxLD_AGxST_sum_f8e5m2(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllReduce_RSxLD_AGxST_sum_f8e5m2 = nullptr;
#endif
#if CUDART_VERSION >= 11080
  __global__ void ncclSymkDevKernel_AllReduce_RSxLD_AGxST_sum_f8e5m2_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllReduce_RSxLD_AGxST_sum_f8e5m2_profile = nullptr;
#endif
#if CUDART_VERSION >= 12070
  __global__ void ncclSymkDevKernel_AllReduce_RSxLDMC_AGxSTMC_sum_f8e5m2(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllReduce_RSxLDMC_AGxSTMC_sum_f8e5m2 = nullptr;
#endif
#if CUDART_VERSION >= 12070
  __global__ void ncclSymkDevKernel_AllReduce_RSxLDMC_AGxSTMC_sum_f8e5m2_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllReduce_RSxLDMC_AGxSTMC_sum_f8e5m2_profile = nullptr;
#endif
#if CUDART_VERSION >= 12000
  __global__ void ncclSymkDevKernel_AllReduce_RSxTmaLD_AGxTmaST_sum_f8e5m2(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllReduce_RSxTmaLD_AGxTmaST_sum_f8e5m2 = nullptr;
#endif
#if CUDART_VERSION >= 12000
  __global__ void ncclSymkDevKernel_AllReduce_RSxTmaLD_AGxTmaST_sum_f8e5m2_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_AllReduce_RSxTmaLD_AGxTmaST_sum_f8e5m2_profile = nullptr;
#endif
#if CUDART_VERSION >= 11080
  __global__ void ncclSymkDevKernel_ReduceScatter_LL_sum_f8e5m2(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LL_sum_f8e5m2 = nullptr;
#endif
#if CUDART_VERSION >= 11080
  __global__ void ncclSymkDevKernel_ReduceScatter_LL_sum_f8e5m2_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LL_sum_f8e5m2_profile = nullptr;
#endif
#if CUDART_VERSION >= 11080
  __global__ void ncclSymkDevKernel_ReduceScatter_LD_sum_f8e5m2(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LD_sum_f8e5m2 = nullptr;
#endif
#if CUDART_VERSION >= 11080
  __global__ void ncclSymkDevKernel_ReduceScatter_LD_sum_f8e5m2_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LD_sum_f8e5m2_profile = nullptr;
#endif
#if CUDART_VERSION >= 12070
  __global__ void ncclSymkDevKernel_ReduceScatter_LDMC_sum_f8e5m2(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LDMC_sum_f8e5m2 = nullptr;
#endif
#if CUDART_VERSION >= 12070
  __global__ void ncclSymkDevKernel_ReduceScatter_LDMC_sum_f8e5m2_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LDMC_sum_f8e5m2_profile = nullptr;
#endif
#if CUDART_VERSION >= 12000
  __global__ void ncclSymkDevKernel_ReduceScatter_TmaLD_sum_f8e5m2(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_TmaLD_sum_f8e5m2 = nullptr;
#endif
#if CUDART_VERSION >= 12000
  __global__ void ncclSymkDevKernel_ReduceScatter_TmaLD_sum_f8e5m2_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_TmaLD_sum_f8e5m2_profile = nullptr;
#endif
#if CUDART_VERSION >= 11080
  __global__ void ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_sum_f8e5m2(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_sum_f8e5m2 = nullptr;
#endif
#if CUDART_VERSION >= 11080
  __global__ void ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_sum_f8e5m2_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_sum_f8e5m2_profile = nullptr;
#endif
#if CUDART_VERSION >= 12070
  __global__ void ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_sum_f8e5m2(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_sum_f8e5m2 = nullptr;
#endif
#if CUDART_VERSION >= 12070
  __global__ void ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_sum_f8e5m2_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_sum_f8e5m2_profile = nullptr;
#endif
#if CUDART_VERSION >= 0
  __global__ void ncclSymkDevKernel_ReduceScatter_LL_avg_f32(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LL_avg_f32 = nullptr;
#endif
#if CUDART_VERSION >= 0
  __global__ void ncclSymkDevKernel_ReduceScatter_LL_avg_f32_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LL_avg_f32_profile = nullptr;
#endif
#if CUDART_VERSION >= 0
  __global__ void ncclSymkDevKernel_ReduceScatter_LD_avg_f32(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LD_avg_f32 = nullptr;
#endif
#if CUDART_VERSION >= 0
  __global__ void ncclSymkDevKernel_ReduceScatter_LD_avg_f32_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LD_avg_f32_profile = nullptr;
#endif
#if CUDART_VERSION >= 12010
  __global__ void ncclSymkDevKernel_ReduceScatter_LDMC_avg_f32(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LDMC_avg_f32 = nullptr;
#endif
#if CUDART_VERSION >= 12010
  __global__ void ncclSymkDevKernel_ReduceScatter_LDMC_avg_f32_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LDMC_avg_f32_profile = nullptr;
#endif
#if CUDART_VERSION >= 12000
  __global__ void ncclSymkDevKernel_ReduceScatter_TmaLD_avg_f32(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_TmaLD_avg_f32 = nullptr;
#endif
#if CUDART_VERSION >= 12000
  __global__ void ncclSymkDevKernel_ReduceScatter_TmaLD_avg_f32_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_TmaLD_avg_f32_profile = nullptr;
#endif
#if CUDART_VERSION >= 0
  __global__ void ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_avg_f32(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_avg_f32 = nullptr;
#endif
#if CUDART_VERSION >= 0
  __global__ void ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_avg_f32_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_avg_f32_profile = nullptr;
#endif
#if CUDART_VERSION >= 12010
  __global__ void ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_avg_f32(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_avg_f32 = nullptr;
#endif
#if CUDART_VERSION >= 12010
  __global__ void ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_avg_f32_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_avg_f32_profile = nullptr;
#endif
#if CUDART_VERSION >= 0
  __global__ void ncclSymkDevKernel_ReduceScatter_LL_avg_f16(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LL_avg_f16 = nullptr;
#endif
#if CUDART_VERSION >= 0
  __global__ void ncclSymkDevKernel_ReduceScatter_LL_avg_f16_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LL_avg_f16_profile = nullptr;
#endif
#if CUDART_VERSION >= 0
  __global__ void ncclSymkDevKernel_ReduceScatter_LD_avg_f16(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LD_avg_f16 = nullptr;
#endif
#if CUDART_VERSION >= 0
  __global__ void ncclSymkDevKernel_ReduceScatter_LD_avg_f16_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LD_avg_f16_profile = nullptr;
#endif
#if CUDART_VERSION >= 12010
  __global__ void ncclSymkDevKernel_ReduceScatter_LDMC_avg_f16(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LDMC_avg_f16 = nullptr;
#endif
#if CUDART_VERSION >= 12010
  __global__ void ncclSymkDevKernel_ReduceScatter_LDMC_avg_f16_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LDMC_avg_f16_profile = nullptr;
#endif
#if CUDART_VERSION >= 12000
  __global__ void ncclSymkDevKernel_ReduceScatter_TmaLD_avg_f16(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_TmaLD_avg_f16 = nullptr;
#endif
#if CUDART_VERSION >= 12000
  __global__ void ncclSymkDevKernel_ReduceScatter_TmaLD_avg_f16_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_TmaLD_avg_f16_profile = nullptr;
#endif
#if CUDART_VERSION >= 0
  __global__ void ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_avg_f16(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_avg_f16 = nullptr;
#endif
#if CUDART_VERSION >= 0
  __global__ void ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_avg_f16_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_avg_f16_profile = nullptr;
#endif
#if CUDART_VERSION >= 12010
  __global__ void ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_avg_f16(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_avg_f16 = nullptr;
#endif
#if CUDART_VERSION >= 12010
  __global__ void ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_avg_f16_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_avg_f16_profile = nullptr;
#endif
#if CUDART_VERSION >= 11000
  __global__ void ncclSymkDevKernel_ReduceScatter_LL_avg_bf16(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LL_avg_bf16 = nullptr;
#endif
#if CUDART_VERSION >= 11000
  __global__ void ncclSymkDevKernel_ReduceScatter_LL_avg_bf16_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LL_avg_bf16_profile = nullptr;
#endif
#if CUDART_VERSION >= 11000
  __global__ void ncclSymkDevKernel_ReduceScatter_LD_avg_bf16(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LD_avg_bf16 = nullptr;
#endif
#if CUDART_VERSION >= 11000
  __global__ void ncclSymkDevKernel_ReduceScatter_LD_avg_bf16_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LD_avg_bf16_profile = nullptr;
#endif
#if CUDART_VERSION >= 12010
  __global__ void ncclSymkDevKernel_ReduceScatter_LDMC_avg_bf16(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LDMC_avg_bf16 = nullptr;
#endif
#if CUDART_VERSION >= 12010
  __global__ void ncclSymkDevKernel_ReduceScatter_LDMC_avg_bf16_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LDMC_avg_bf16_profile = nullptr;
#endif
#if CUDART_VERSION >= 12000
  __global__ void ncclSymkDevKernel_ReduceScatter_TmaLD_avg_bf16(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_TmaLD_avg_bf16 = nullptr;
#endif
#if CUDART_VERSION >= 12000
  __global__ void ncclSymkDevKernel_ReduceScatter_TmaLD_avg_bf16_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_TmaLD_avg_bf16_profile = nullptr;
#endif
#if CUDART_VERSION >= 11000
  __global__ void ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_avg_bf16(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_avg_bf16 = nullptr;
#endif
#if CUDART_VERSION >= 11000
  __global__ void ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_avg_bf16_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_avg_bf16_profile = nullptr;
#endif
#if CUDART_VERSION >= 12010
  __global__ void ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_avg_bf16(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_avg_bf16 = nullptr;
#endif
#if CUDART_VERSION >= 12010
  __global__ void ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_avg_bf16_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_avg_bf16_profile = nullptr;
#endif
#if CUDART_VERSION >= 11080
  __global__ void ncclSymkDevKernel_ReduceScatter_LL_avg_f8e4m3(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LL_avg_f8e4m3 = nullptr;
#endif
#if CUDART_VERSION >= 11080
  __global__ void ncclSymkDevKernel_ReduceScatter_LL_avg_f8e4m3_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LL_avg_f8e4m3_profile = nullptr;
#endif
#if CUDART_VERSION >= 11080
  __global__ void ncclSymkDevKernel_ReduceScatter_LD_avg_f8e4m3(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LD_avg_f8e4m3 = nullptr;
#endif
#if CUDART_VERSION >= 11080
  __global__ void ncclSymkDevKernel_ReduceScatter_LD_avg_f8e4m3_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LD_avg_f8e4m3_profile = nullptr;
#endif
#if CUDART_VERSION >= 12070
  __global__ void ncclSymkDevKernel_ReduceScatter_LDMC_avg_f8e4m3(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LDMC_avg_f8e4m3 = nullptr;
#endif
#if CUDART_VERSION >= 12070
  __global__ void ncclSymkDevKernel_ReduceScatter_LDMC_avg_f8e4m3_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LDMC_avg_f8e4m3_profile = nullptr;
#endif
#if CUDART_VERSION >= 12000
  __global__ void ncclSymkDevKernel_ReduceScatter_TmaLD_avg_f8e4m3(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_TmaLD_avg_f8e4m3 = nullptr;
#endif
#if CUDART_VERSION >= 12000
  __global__ void ncclSymkDevKernel_ReduceScatter_TmaLD_avg_f8e4m3_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_TmaLD_avg_f8e4m3_profile = nullptr;
#endif
#if CUDART_VERSION >= 11080
  __global__ void ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_avg_f8e4m3(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_avg_f8e4m3 = nullptr;
#endif
#if CUDART_VERSION >= 11080
  __global__ void ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_avg_f8e4m3_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_avg_f8e4m3_profile = nullptr;
#endif
#if CUDART_VERSION >= 12070
  __global__ void ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_avg_f8e4m3(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_avg_f8e4m3 = nullptr;
#endif
#if CUDART_VERSION >= 12070
  __global__ void ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_avg_f8e4m3_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_avg_f8e4m3_profile = nullptr;
#endif
#if CUDART_VERSION >= 11080
  __global__ void ncclSymkDevKernel_ReduceScatter_LL_avg_f8e5m2(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LL_avg_f8e5m2 = nullptr;
#endif
#if CUDART_VERSION >= 11080
  __global__ void ncclSymkDevKernel_ReduceScatter_LL_avg_f8e5m2_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LL_avg_f8e5m2_profile = nullptr;
#endif
#if CUDART_VERSION >= 11080
  __global__ void ncclSymkDevKernel_ReduceScatter_LD_avg_f8e5m2(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LD_avg_f8e5m2 = nullptr;
#endif
#if CUDART_VERSION >= 11080
  __global__ void ncclSymkDevKernel_ReduceScatter_LD_avg_f8e5m2_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LD_avg_f8e5m2_profile = nullptr;
#endif
#if CUDART_VERSION >= 12070
  __global__ void ncclSymkDevKernel_ReduceScatter_LDMC_avg_f8e5m2(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LDMC_avg_f8e5m2 = nullptr;
#endif
#if CUDART_VERSION >= 12070
  __global__ void ncclSymkDevKernel_ReduceScatter_LDMC_avg_f8e5m2_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_LDMC_avg_f8e5m2_profile = nullptr;
#endif
#if CUDART_VERSION >= 12000
  __global__ void ncclSymkDevKernel_ReduceScatter_TmaLD_avg_f8e5m2(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_TmaLD_avg_f8e5m2 = nullptr;
#endif
#if CUDART_VERSION >= 12000
  __global__ void ncclSymkDevKernel_ReduceScatter_TmaLD_avg_f8e5m2_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_TmaLD_avg_f8e5m2_profile = nullptr;
#endif
#if CUDART_VERSION >= 11080
  __global__ void ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_avg_f8e5m2(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_avg_f8e5m2 = nullptr;
#endif
#if CUDART_VERSION >= 11080
  __global__ void ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_avg_f8e5m2_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_avg_f8e5m2_profile = nullptr;
#endif
#if CUDART_VERSION >= 12070
  __global__ void ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_avg_f8e5m2(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_avg_f8e5m2 = nullptr;
#endif
#if CUDART_VERSION >= 12070
  __global__ void ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_avg_f8e5m2_profile(ncclSymkDevWorkArgs4K const);
#else
  constexpr void* ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_avg_f8e5m2_profile = nullptr;
#endif

extern int const ncclSymkKernelCount = 92;
void* ncclSymkKernelList[] = {
(void*)ncclSymkDevKernel_AllGather_LL,
(void*)ncclSymkDevKernel_AllGather_LLMC,
(void*)ncclSymkDevKernel_AllGather_ST,
(void*)ncclSymkDevKernel_AllGather_STMC,
(void*)ncclSymkDevKernel_AllGather_TmaST,
(void*)ncclSymkDevKernel_AllGather_TmaSTMC,
(void*)ncclSymkDevKernel_AllGather_RailRing_LsaSTMC,
(void*)ncclSymkDevKernel_AllReduce_AGxLL_R_sum_f32,
(void*)ncclSymkDevKernel_AllReduce_AGxLLMC_R_sum_f32,
(void*)ncclSymkDevKernel_AllReduce_RSxLD_AGxST_sum_f32,
(void*)ncclSymkDevKernel_AllReduce_RSxLDMC_AGxSTMC_sum_f32,
(void*)ncclSymkDevKernel_AllReduce_RSxTmaLD_AGxTmaST_sum_f32,
(void*)ncclSymkDevKernel_ReduceScatter_LL_sum_f32,
(void*)ncclSymkDevKernel_ReduceScatter_LD_sum_f32,
(void*)ncclSymkDevKernel_ReduceScatter_LDMC_sum_f32,
(void*)ncclSymkDevKernel_ReduceScatter_TmaLD_sum_f32,
(void*)ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_sum_f32,
(void*)ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_sum_f32,
(void*)ncclSymkDevKernel_AllReduce_AGxLL_R_sum_f16,
(void*)ncclSymkDevKernel_AllReduce_AGxLLMC_R_sum_f16,
(void*)ncclSymkDevKernel_AllReduce_RSxLD_AGxST_sum_f16,
(void*)ncclSymkDevKernel_AllReduce_RSxLDMC_AGxSTMC_sum_f16,
(void*)ncclSymkDevKernel_AllReduce_RSxTmaLD_AGxTmaST_sum_f16,
(void*)ncclSymkDevKernel_ReduceScatter_LL_sum_f16,
(void*)ncclSymkDevKernel_ReduceScatter_LD_sum_f16,
(void*)ncclSymkDevKernel_ReduceScatter_LDMC_sum_f16,
(void*)ncclSymkDevKernel_ReduceScatter_TmaLD_sum_f16,
(void*)ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_sum_f16,
(void*)ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_sum_f16,
(void*)ncclSymkDevKernel_AllReduce_AGxLL_R_sum_bf16,
(void*)ncclSymkDevKernel_AllReduce_AGxLLMC_R_sum_bf16,
(void*)ncclSymkDevKernel_AllReduce_RSxLD_AGxST_sum_bf16,
(void*)ncclSymkDevKernel_AllReduce_RSxLDMC_AGxSTMC_sum_bf16,
(void*)ncclSymkDevKernel_AllReduce_RSxTmaLD_AGxTmaST_sum_bf16,
(void*)ncclSymkDevKernel_ReduceScatter_LL_sum_bf16,
(void*)ncclSymkDevKernel_ReduceScatter_LD_sum_bf16,
(void*)ncclSymkDevKernel_ReduceScatter_LDMC_sum_bf16,
(void*)ncclSymkDevKernel_ReduceScatter_TmaLD_sum_bf16,
(void*)ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_sum_bf16,
(void*)ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_sum_bf16,
(void*)ncclSymkDevKernel_AllReduce_AGxLL_R_sum_f8e4m3,
(void*)ncclSymkDevKernel_AllReduce_AGxLLMC_R_sum_f8e4m3,
(void*)ncclSymkDevKernel_AllReduce_RSxLD_AGxST_sum_f8e4m3,
(void*)ncclSymkDevKernel_AllReduce_RSxLDMC_AGxSTMC_sum_f8e4m3,
(void*)ncclSymkDevKernel_AllReduce_RSxTmaLD_AGxTmaST_sum_f8e4m3,
(void*)ncclSymkDevKernel_ReduceScatter_LL_sum_f8e4m3,
(void*)ncclSymkDevKernel_ReduceScatter_LD_sum_f8e4m3,
(void*)ncclSymkDevKernel_ReduceScatter_LDMC_sum_f8e4m3,
(void*)ncclSymkDevKernel_ReduceScatter_TmaLD_sum_f8e4m3,
(void*)ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_sum_f8e4m3,
(void*)ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_sum_f8e4m3,
(void*)ncclSymkDevKernel_AllReduce_AGxLL_R_sum_f8e5m2,
(void*)ncclSymkDevKernel_AllReduce_AGxLLMC_R_sum_f8e5m2,
(void*)ncclSymkDevKernel_AllReduce_RSxLD_AGxST_sum_f8e5m2,
(void*)ncclSymkDevKernel_AllReduce_RSxLDMC_AGxSTMC_sum_f8e5m2,
(void*)ncclSymkDevKernel_AllReduce_RSxTmaLD_AGxTmaST_sum_f8e5m2,
(void*)ncclSymkDevKernel_ReduceScatter_LL_sum_f8e5m2,
(void*)ncclSymkDevKernel_ReduceScatter_LD_sum_f8e5m2,
(void*)ncclSymkDevKernel_ReduceScatter_LDMC_sum_f8e5m2,
(void*)ncclSymkDevKernel_ReduceScatter_TmaLD_sum_f8e5m2,
(void*)ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_sum_f8e5m2,
(void*)ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_sum_f8e5m2,
(void*)ncclSymkDevKernel_ReduceScatter_LL_avg_f32,
(void*)ncclSymkDevKernel_ReduceScatter_LD_avg_f32,
(void*)ncclSymkDevKernel_ReduceScatter_LDMC_avg_f32,
(void*)ncclSymkDevKernel_ReduceScatter_TmaLD_avg_f32,
(void*)ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_avg_f32,
(void*)ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_avg_f32,
(void*)ncclSymkDevKernel_ReduceScatter_LL_avg_f16,
(void*)ncclSymkDevKernel_ReduceScatter_LD_avg_f16,
(void*)ncclSymkDevKernel_ReduceScatter_LDMC_avg_f16,
(void*)ncclSymkDevKernel_ReduceScatter_TmaLD_avg_f16,
(void*)ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_avg_f16,
(void*)ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_avg_f16,
(void*)ncclSymkDevKernel_ReduceScatter_LL_avg_bf16,
(void*)ncclSymkDevKernel_ReduceScatter_LD_avg_bf16,
(void*)ncclSymkDevKernel_ReduceScatter_LDMC_avg_bf16,
(void*)ncclSymkDevKernel_ReduceScatter_TmaLD_avg_bf16,
(void*)ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_avg_bf16,
(void*)ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_avg_bf16,
(void*)ncclSymkDevKernel_ReduceScatter_LL_avg_f8e4m3,
(void*)ncclSymkDevKernel_ReduceScatter_LD_avg_f8e4m3,
(void*)ncclSymkDevKernel_ReduceScatter_LDMC_avg_f8e4m3,
(void*)ncclSymkDevKernel_ReduceScatter_TmaLD_avg_f8e4m3,
(void*)ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_avg_f8e4m3,
(void*)ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_avg_f8e4m3,
(void*)ncclSymkDevKernel_ReduceScatter_LL_avg_f8e5m2,
(void*)ncclSymkDevKernel_ReduceScatter_LD_avg_f8e5m2,
(void*)ncclSymkDevKernel_ReduceScatter_LDMC_avg_f8e5m2,
(void*)ncclSymkDevKernel_ReduceScatter_TmaLD_avg_f8e5m2,
(void*)ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_avg_f8e5m2,
(void*)ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_avg_f8e5m2,
nullptr};

void* ncclSymkKernelListProfile[] = {
(void*)ncclSymkDevKernel_AllGather_LL_profile,
(void*)ncclSymkDevKernel_AllGather_LLMC_profile,
(void*)ncclSymkDevKernel_AllGather_ST_profile,
(void*)ncclSymkDevKernel_AllGather_STMC_profile,
(void*)ncclSymkDevKernel_AllGather_TmaST_profile,
(void*)ncclSymkDevKernel_AllGather_TmaSTMC_profile,
(void*)ncclSymkDevKernel_AllGather_RailRing_LsaSTMC_profile,
(void*)ncclSymkDevKernel_AllReduce_AGxLL_R_sum_f32_profile,
(void*)ncclSymkDevKernel_AllReduce_AGxLLMC_R_sum_f32_profile,
(void*)ncclSymkDevKernel_AllReduce_RSxLD_AGxST_sum_f32_profile,
(void*)ncclSymkDevKernel_AllReduce_RSxLDMC_AGxSTMC_sum_f32_profile,
(void*)ncclSymkDevKernel_AllReduce_RSxTmaLD_AGxTmaST_sum_f32_profile,
(void*)ncclSymkDevKernel_ReduceScatter_LL_sum_f32_profile,
(void*)ncclSymkDevKernel_ReduceScatter_LD_sum_f32_profile,
(void*)ncclSymkDevKernel_ReduceScatter_LDMC_sum_f32_profile,
(void*)ncclSymkDevKernel_ReduceScatter_TmaLD_sum_f32_profile,
(void*)ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_sum_f32_profile,
(void*)ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_sum_f32_profile,
(void*)ncclSymkDevKernel_AllReduce_AGxLL_R_sum_f16_profile,
(void*)ncclSymkDevKernel_AllReduce_AGxLLMC_R_sum_f16_profile,
(void*)ncclSymkDevKernel_AllReduce_RSxLD_AGxST_sum_f16_profile,
(void*)ncclSymkDevKernel_AllReduce_RSxLDMC_AGxSTMC_sum_f16_profile,
(void*)ncclSymkDevKernel_AllReduce_RSxTmaLD_AGxTmaST_sum_f16_profile,
(void*)ncclSymkDevKernel_ReduceScatter_LL_sum_f16_profile,
(void*)ncclSymkDevKernel_ReduceScatter_LD_sum_f16_profile,
(void*)ncclSymkDevKernel_ReduceScatter_LDMC_sum_f16_profile,
(void*)ncclSymkDevKernel_ReduceScatter_TmaLD_sum_f16_profile,
(void*)ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_sum_f16_profile,
(void*)ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_sum_f16_profile,
(void*)ncclSymkDevKernel_AllReduce_AGxLL_R_sum_bf16_profile,
(void*)ncclSymkDevKernel_AllReduce_AGxLLMC_R_sum_bf16_profile,
(void*)ncclSymkDevKernel_AllReduce_RSxLD_AGxST_sum_bf16_profile,
(void*)ncclSymkDevKernel_AllReduce_RSxLDMC_AGxSTMC_sum_bf16_profile,
(void*)ncclSymkDevKernel_AllReduce_RSxTmaLD_AGxTmaST_sum_bf16_profile,
(void*)ncclSymkDevKernel_ReduceScatter_LL_sum_bf16_profile,
(void*)ncclSymkDevKernel_ReduceScatter_LD_sum_bf16_profile,
(void*)ncclSymkDevKernel_ReduceScatter_LDMC_sum_bf16_profile,
(void*)ncclSymkDevKernel_ReduceScatter_TmaLD_sum_bf16_profile,
(void*)ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_sum_bf16_profile,
(void*)ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_sum_bf16_profile,
(void*)ncclSymkDevKernel_AllReduce_AGxLL_R_sum_f8e4m3_profile,
(void*)ncclSymkDevKernel_AllReduce_AGxLLMC_R_sum_f8e4m3_profile,
(void*)ncclSymkDevKernel_AllReduce_RSxLD_AGxST_sum_f8e4m3_profile,
(void*)ncclSymkDevKernel_AllReduce_RSxLDMC_AGxSTMC_sum_f8e4m3_profile,
(void*)ncclSymkDevKernel_AllReduce_RSxTmaLD_AGxTmaST_sum_f8e4m3_profile,
(void*)ncclSymkDevKernel_ReduceScatter_LL_sum_f8e4m3_profile,
(void*)ncclSymkDevKernel_ReduceScatter_LD_sum_f8e4m3_profile,
(void*)ncclSymkDevKernel_ReduceScatter_LDMC_sum_f8e4m3_profile,
(void*)ncclSymkDevKernel_ReduceScatter_TmaLD_sum_f8e4m3_profile,
(void*)ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_sum_f8e4m3_profile,
(void*)ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_sum_f8e4m3_profile,
(void*)ncclSymkDevKernel_AllReduce_AGxLL_R_sum_f8e5m2_profile,
(void*)ncclSymkDevKernel_AllReduce_AGxLLMC_R_sum_f8e5m2_profile,
(void*)ncclSymkDevKernel_AllReduce_RSxLD_AGxST_sum_f8e5m2_profile,
(void*)ncclSymkDevKernel_AllReduce_RSxLDMC_AGxSTMC_sum_f8e5m2_profile,
(void*)ncclSymkDevKernel_AllReduce_RSxTmaLD_AGxTmaST_sum_f8e5m2_profile,
(void*)ncclSymkDevKernel_ReduceScatter_LL_sum_f8e5m2_profile,
(void*)ncclSymkDevKernel_ReduceScatter_LD_sum_f8e5m2_profile,
(void*)ncclSymkDevKernel_ReduceScatter_LDMC_sum_f8e5m2_profile,
(void*)ncclSymkDevKernel_ReduceScatter_TmaLD_sum_f8e5m2_profile,
(void*)ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_sum_f8e5m2_profile,
(void*)ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_sum_f8e5m2_profile,
(void*)ncclSymkDevKernel_ReduceScatter_LL_avg_f32_profile,
(void*)ncclSymkDevKernel_ReduceScatter_LD_avg_f32_profile,
(void*)ncclSymkDevKernel_ReduceScatter_LDMC_avg_f32_profile,
(void*)ncclSymkDevKernel_ReduceScatter_TmaLD_avg_f32_profile,
(void*)ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_avg_f32_profile,
(void*)ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_avg_f32_profile,
(void*)ncclSymkDevKernel_ReduceScatter_LL_avg_f16_profile,
(void*)ncclSymkDevKernel_ReduceScatter_LD_avg_f16_profile,
(void*)ncclSymkDevKernel_ReduceScatter_LDMC_avg_f16_profile,
(void*)ncclSymkDevKernel_ReduceScatter_TmaLD_avg_f16_profile,
(void*)ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_avg_f16_profile,
(void*)ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_avg_f16_profile,
(void*)ncclSymkDevKernel_ReduceScatter_LL_avg_bf16_profile,
(void*)ncclSymkDevKernel_ReduceScatter_LD_avg_bf16_profile,
(void*)ncclSymkDevKernel_ReduceScatter_LDMC_avg_bf16_profile,
(void*)ncclSymkDevKernel_ReduceScatter_TmaLD_avg_bf16_profile,
(void*)ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_avg_bf16_profile,
(void*)ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_avg_bf16_profile,
(void*)ncclSymkDevKernel_ReduceScatter_LL_avg_f8e4m3_profile,
(void*)ncclSymkDevKernel_ReduceScatter_LD_avg_f8e4m3_profile,
(void*)ncclSymkDevKernel_ReduceScatter_LDMC_avg_f8e4m3_profile,
(void*)ncclSymkDevKernel_ReduceScatter_TmaLD_avg_f8e4m3_profile,
(void*)ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_avg_f8e4m3_profile,
(void*)ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_avg_f8e4m3_profile,
(void*)ncclSymkDevKernel_ReduceScatter_LL_avg_f8e5m2_profile,
(void*)ncclSymkDevKernel_ReduceScatter_LD_avg_f8e5m2_profile,
(void*)ncclSymkDevKernel_ReduceScatter_LDMC_avg_f8e5m2_profile,
(void*)ncclSymkDevKernel_ReduceScatter_TmaLD_avg_f8e5m2_profile,
(void*)ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_avg_f8e5m2_profile,
(void*)ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_avg_f8e5m2_profile,
nullptr};

int ncclSymkKernelRequirements[] = {
        0, /*   0 ncclSymkDevKernel_AllGather_LL*/
        0, /*   1 ncclSymkDevKernel_AllGather_LLMC*/
        0, /*   2 ncclSymkDevKernel_AllGather_ST*/
        0, /*   3 ncclSymkDevKernel_AllGather_STMC*/
    12000, /*   4 ncclSymkDevKernel_AllGather_TmaST*/
    12000, /*   5 ncclSymkDevKernel_AllGather_TmaSTMC*/
        0, /*   6 ncclSymkDevKernel_AllGather_RailRing_LsaSTMC*/
        0, /*   7 ncclSymkDevKernel_AllReduce_AGxLL_R_sum_f32*/
    12010, /*   8 ncclSymkDevKernel_AllReduce_AGxLLMC_R_sum_f32*/
        0, /*   9 ncclSymkDevKernel_AllReduce_RSxLD_AGxST_sum_f32*/
    12010, /*  10 ncclSymkDevKernel_AllReduce_RSxLDMC_AGxSTMC_sum_f32*/
    12000, /*  11 ncclSymkDevKernel_AllReduce_RSxTmaLD_AGxTmaST_sum_f32*/
        0, /*  12 ncclSymkDevKernel_ReduceScatter_LL_sum_f32*/
        0, /*  13 ncclSymkDevKernel_ReduceScatter_LD_sum_f32*/
    12010, /*  14 ncclSymkDevKernel_ReduceScatter_LDMC_sum_f32*/
    12000, /*  15 ncclSymkDevKernel_ReduceScatter_TmaLD_sum_f32*/
        0, /*  16 ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_sum_f32*/
    12010, /*  17 ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_sum_f32*/
        0, /*  18 ncclSymkDevKernel_AllReduce_AGxLL_R_sum_f16*/
    12010, /*  19 ncclSymkDevKernel_AllReduce_AGxLLMC_R_sum_f16*/
        0, /*  20 ncclSymkDevKernel_AllReduce_RSxLD_AGxST_sum_f16*/
    12010, /*  21 ncclSymkDevKernel_AllReduce_RSxLDMC_AGxSTMC_sum_f16*/
    12000, /*  22 ncclSymkDevKernel_AllReduce_RSxTmaLD_AGxTmaST_sum_f16*/
        0, /*  23 ncclSymkDevKernel_ReduceScatter_LL_sum_f16*/
        0, /*  24 ncclSymkDevKernel_ReduceScatter_LD_sum_f16*/
    12010, /*  25 ncclSymkDevKernel_ReduceScatter_LDMC_sum_f16*/
    12000, /*  26 ncclSymkDevKernel_ReduceScatter_TmaLD_sum_f16*/
        0, /*  27 ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_sum_f16*/
    12010, /*  28 ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_sum_f16*/
    11000, /*  29 ncclSymkDevKernel_AllReduce_AGxLL_R_sum_bf16*/
    12010, /*  30 ncclSymkDevKernel_AllReduce_AGxLLMC_R_sum_bf16*/
    11000, /*  31 ncclSymkDevKernel_AllReduce_RSxLD_AGxST_sum_bf16*/
    12010, /*  32 ncclSymkDevKernel_AllReduce_RSxLDMC_AGxSTMC_sum_bf16*/
    12000, /*  33 ncclSymkDevKernel_AllReduce_RSxTmaLD_AGxTmaST_sum_bf16*/
    11000, /*  34 ncclSymkDevKernel_ReduceScatter_LL_sum_bf16*/
    11000, /*  35 ncclSymkDevKernel_ReduceScatter_LD_sum_bf16*/
    12010, /*  36 ncclSymkDevKernel_ReduceScatter_LDMC_sum_bf16*/
    12000, /*  37 ncclSymkDevKernel_ReduceScatter_TmaLD_sum_bf16*/
    11000, /*  38 ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_sum_bf16*/
    12010, /*  39 ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_sum_bf16*/
    11080, /*  40 ncclSymkDevKernel_AllReduce_AGxLL_R_sum_f8e4m3*/
    12010, /*  41 ncclSymkDevKernel_AllReduce_AGxLLMC_R_sum_f8e4m3*/
    11080, /*  42 ncclSymkDevKernel_AllReduce_RSxLD_AGxST_sum_f8e4m3*/
    12070, /*  43 ncclSymkDevKernel_AllReduce_RSxLDMC_AGxSTMC_sum_f8e4m3*/
    12000, /*  44 ncclSymkDevKernel_AllReduce_RSxTmaLD_AGxTmaST_sum_f8e4m3*/
    11080, /*  45 ncclSymkDevKernel_ReduceScatter_LL_sum_f8e4m3*/
    11080, /*  46 ncclSymkDevKernel_ReduceScatter_LD_sum_f8e4m3*/
    12070, /*  47 ncclSymkDevKernel_ReduceScatter_LDMC_sum_f8e4m3*/
    12000, /*  48 ncclSymkDevKernel_ReduceScatter_TmaLD_sum_f8e4m3*/
    11080, /*  49 ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_sum_f8e4m3*/
    12070, /*  50 ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_sum_f8e4m3*/
    11080, /*  51 ncclSymkDevKernel_AllReduce_AGxLL_R_sum_f8e5m2*/
    12010, /*  52 ncclSymkDevKernel_AllReduce_AGxLLMC_R_sum_f8e5m2*/
    11080, /*  53 ncclSymkDevKernel_AllReduce_RSxLD_AGxST_sum_f8e5m2*/
    12070, /*  54 ncclSymkDevKernel_AllReduce_RSxLDMC_AGxSTMC_sum_f8e5m2*/
    12000, /*  55 ncclSymkDevKernel_AllReduce_RSxTmaLD_AGxTmaST_sum_f8e5m2*/
    11080, /*  56 ncclSymkDevKernel_ReduceScatter_LL_sum_f8e5m2*/
    11080, /*  57 ncclSymkDevKernel_ReduceScatter_LD_sum_f8e5m2*/
    12070, /*  58 ncclSymkDevKernel_ReduceScatter_LDMC_sum_f8e5m2*/
    12000, /*  59 ncclSymkDevKernel_ReduceScatter_TmaLD_sum_f8e5m2*/
    11080, /*  60 ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_sum_f8e5m2*/
    12070, /*  61 ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_sum_f8e5m2*/
        0, /*  62 ncclSymkDevKernel_ReduceScatter_LL_avg_f32*/
        0, /*  63 ncclSymkDevKernel_ReduceScatter_LD_avg_f32*/
    12010, /*  64 ncclSymkDevKernel_ReduceScatter_LDMC_avg_f32*/
    12000, /*  65 ncclSymkDevKernel_ReduceScatter_TmaLD_avg_f32*/
        0, /*  66 ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_avg_f32*/
    12010, /*  67 ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_avg_f32*/
        0, /*  68 ncclSymkDevKernel_ReduceScatter_LL_avg_f16*/
        0, /*  69 ncclSymkDevKernel_ReduceScatter_LD_avg_f16*/
    12010, /*  70 ncclSymkDevKernel_ReduceScatter_LDMC_avg_f16*/
    12000, /*  71 ncclSymkDevKernel_ReduceScatter_TmaLD_avg_f16*/
        0, /*  72 ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_avg_f16*/
    12010, /*  73 ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_avg_f16*/
    11000, /*  74 ncclSymkDevKernel_ReduceScatter_LL_avg_bf16*/
    11000, /*  75 ncclSymkDevKernel_ReduceScatter_LD_avg_bf16*/
    12010, /*  76 ncclSymkDevKernel_ReduceScatter_LDMC_avg_bf16*/
    12000, /*  77 ncclSymkDevKernel_ReduceScatter_TmaLD_avg_bf16*/
    11000, /*  78 ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_avg_bf16*/
    12010, /*  79 ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_avg_bf16*/
    11080, /*  80 ncclSymkDevKernel_ReduceScatter_LL_avg_f8e4m3*/
    11080, /*  81 ncclSymkDevKernel_ReduceScatter_LD_avg_f8e4m3*/
    12070, /*  82 ncclSymkDevKernel_ReduceScatter_LDMC_avg_f8e4m3*/
    12000, /*  83 ncclSymkDevKernel_ReduceScatter_TmaLD_avg_f8e4m3*/
    11080, /*  84 ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_avg_f8e4m3*/
    12070, /*  85 ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_avg_f8e4m3*/
    11080, /*  86 ncclSymkDevKernel_ReduceScatter_LL_avg_f8e5m2*/
    11080, /*  87 ncclSymkDevKernel_ReduceScatter_LD_avg_f8e5m2*/
    12070, /*  88 ncclSymkDevKernel_ReduceScatter_LDMC_avg_f8e5m2*/
    12000, /*  89 ncclSymkDevKernel_ReduceScatter_TmaLD_avg_f8e5m2*/
    11080, /*  90 ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLD_avg_f8e5m2*/
    12070, /*  91 ncclSymkDevKernel_ReduceScatter_RailA2A_LsaLDMC_avg_f8e5m2*/
};

int ncclSymkKernelMaxDynamicSmem[92];

int ncclSymkGetKernelIndex(ncclSymkKernelId id, int red, ncclDataType_t ty) {
  switch (id) {
  default: WARN("ncclSymkGetKernelIndex: unknown kernel id %d", (int)id); return -1;
  case ncclSymkKernelId_AllGather_LL:
    return 0;
  case ncclSymkKernelId_AllGather_LLMC:
    return 1;
  case ncclSymkKernelId_AllGather_ST:
    return 2;
  case ncclSymkKernelId_AllGather_STMC:
    return 3;
  case ncclSymkKernelId_AllGather_TmaST:
    return 4;
  case ncclSymkKernelId_AllGather_TmaSTMC:
    return 5;
  case ncclSymkKernelId_AllGather_RailRing_LsaSTMC:
    return 6;
  case ncclSymkKernelId_AllReduce_AGxLL_R:
    switch ((ncclDevRedOp_t)red) {
    default: WARN("ncclSymkGetKernelIndex: unknown red op %d for id %d", red, (int)id); return -1;
    case ncclDevSum:
      switch (ty) {
      default: WARN("ncclSymkGetKernelIndex: unknown type %d for id %d red %d", (int)ty, (int)id, red); return -1;
      case ncclFloat32: return 7;
      case ncclFloat16: return 18;
      case ncclBfloat16: return 29;
      case ncclFloat8e4m3: return 40;
      case ncclFloat8e5m2: return 51;
      }
    }
  case ncclSymkKernelId_AllReduce_AGxLLMC_R:
    switch ((ncclDevRedOp_t)red) {
    default: WARN("ncclSymkGetKernelIndex: unknown red op %d for id %d", red, (int)id); return -1;
    case ncclDevSum:
      switch (ty) {
      default: WARN("ncclSymkGetKernelIndex: unknown type %d for id %d red %d", (int)ty, (int)id, red); return -1;
      case ncclFloat32: return 8;
      case ncclFloat16: return 19;
      case ncclBfloat16: return 30;
      case ncclFloat8e4m3: return 41;
      case ncclFloat8e5m2: return 52;
      }
    }
  case ncclSymkKernelId_AllReduce_RSxLD_AGxST:
    switch ((ncclDevRedOp_t)red) {
    default: WARN("ncclSymkGetKernelIndex: unknown red op %d for id %d", red, (int)id); return -1;
    case ncclDevSum:
      switch (ty) {
      default: WARN("ncclSymkGetKernelIndex: unknown type %d for id %d red %d", (int)ty, (int)id, red); return -1;
      case ncclFloat32: return 9;
      case ncclFloat16: return 20;
      case ncclBfloat16: return 31;
      case ncclFloat8e4m3: return 42;
      case ncclFloat8e5m2: return 53;
      }
    }
  case ncclSymkKernelId_AllReduce_RSxLDMC_AGxSTMC:
    switch ((ncclDevRedOp_t)red) {
    default: WARN("ncclSymkGetKernelIndex: unknown red op %d for id %d", red, (int)id); return -1;
    case ncclDevSum:
      switch (ty) {
      default: WARN("ncclSymkGetKernelIndex: unknown type %d for id %d red %d", (int)ty, (int)id, red); return -1;
      case ncclFloat32: return 10;
      case ncclFloat16: return 21;
      case ncclBfloat16: return 32;
      case ncclFloat8e4m3: return 43;
      case ncclFloat8e5m2: return 54;
      }
    }
  case ncclSymkKernelId_AllReduce_RSxTmaLD_AGxTmaST:
    switch ((ncclDevRedOp_t)red) {
    default: WARN("ncclSymkGetKernelIndex: unknown red op %d for id %d", red, (int)id); return -1;
    case ncclDevSum:
      switch (ty) {
      default: WARN("ncclSymkGetKernelIndex: unknown type %d for id %d red %d", (int)ty, (int)id, red); return -1;
      case ncclFloat32: return 11;
      case ncclFloat16: return 22;
      case ncclBfloat16: return 33;
      case ncclFloat8e4m3: return 44;
      case ncclFloat8e5m2: return 55;
      }
    }
  case ncclSymkKernelId_ReduceScatter_LL:
    switch ((ncclDevRedOp_t)red) {
    default: WARN("ncclSymkGetKernelIndex: unknown red op %d for id %d", red, (int)id); return -1;
    case ncclDevSum:
      switch (ty) {
      default: WARN("ncclSymkGetKernelIndex: unknown type %d for id %d red %d", (int)ty, (int)id, red); return -1;
      case ncclFloat32: return 12;
      case ncclFloat16: return 23;
      case ncclBfloat16: return 34;
      case ncclFloat8e4m3: return 45;
      case ncclFloat8e5m2: return 56;
      }
    case ncclDevSumPostDiv:
      switch (ty) {
      default: WARN("ncclSymkGetKernelIndex: unknown type %d for id %d red %d", (int)ty, (int)id, red); return -1;
      case ncclFloat32: return 62;
      case ncclFloat16: return 68;
      case ncclBfloat16: return 74;
      case ncclFloat8e4m3: return 80;
      case ncclFloat8e5m2: return 86;
      }
    }
  case ncclSymkKernelId_ReduceScatter_LD:
    switch ((ncclDevRedOp_t)red) {
    default: WARN("ncclSymkGetKernelIndex: unknown red op %d for id %d", red, (int)id); return -1;
    case ncclDevSum:
      switch (ty) {
      default: WARN("ncclSymkGetKernelIndex: unknown type %d for id %d red %d", (int)ty, (int)id, red); return -1;
      case ncclFloat32: return 13;
      case ncclFloat16: return 24;
      case ncclBfloat16: return 35;
      case ncclFloat8e4m3: return 46;
      case ncclFloat8e5m2: return 57;
      }
    case ncclDevSumPostDiv:
      switch (ty) {
      default: WARN("ncclSymkGetKernelIndex: unknown type %d for id %d red %d", (int)ty, (int)id, red); return -1;
      case ncclFloat32: return 63;
      case ncclFloat16: return 69;
      case ncclBfloat16: return 75;
      case ncclFloat8e4m3: return 81;
      case ncclFloat8e5m2: return 87;
      }
    }
  case ncclSymkKernelId_ReduceScatter_LDMC:
    switch ((ncclDevRedOp_t)red) {
    default: WARN("ncclSymkGetKernelIndex: unknown red op %d for id %d", red, (int)id); return -1;
    case ncclDevSum:
      switch (ty) {
      default: WARN("ncclSymkGetKernelIndex: unknown type %d for id %d red %d", (int)ty, (int)id, red); return -1;
      case ncclFloat32: return 14;
      case ncclFloat16: return 25;
      case ncclBfloat16: return 36;
      case ncclFloat8e4m3: return 47;
      case ncclFloat8e5m2: return 58;
      }
    case ncclDevSumPostDiv:
      switch (ty) {
      default: WARN("ncclSymkGetKernelIndex: unknown type %d for id %d red %d", (int)ty, (int)id, red); return -1;
      case ncclFloat32: return 64;
      case ncclFloat16: return 70;
      case ncclBfloat16: return 76;
      case ncclFloat8e4m3: return 82;
      case ncclFloat8e5m2: return 88;
      }
    }
  case ncclSymkKernelId_ReduceScatter_TmaLD:
    switch ((ncclDevRedOp_t)red) {
    default: WARN("ncclSymkGetKernelIndex: unknown red op %d for id %d", red, (int)id); return -1;
    case ncclDevSum:
      switch (ty) {
      default: WARN("ncclSymkGetKernelIndex: unknown type %d for id %d red %d", (int)ty, (int)id, red); return -1;
      case ncclFloat32: return 15;
      case ncclFloat16: return 26;
      case ncclBfloat16: return 37;
      case ncclFloat8e4m3: return 48;
      case ncclFloat8e5m2: return 59;
      }
    case ncclDevSumPostDiv:
      switch (ty) {
      default: WARN("ncclSymkGetKernelIndex: unknown type %d for id %d red %d", (int)ty, (int)id, red); return -1;
      case ncclFloat32: return 65;
      case ncclFloat16: return 71;
      case ncclBfloat16: return 77;
      case ncclFloat8e4m3: return 83;
      case ncclFloat8e5m2: return 89;
      }
    }
  case ncclSymkKernelId_ReduceScatter_RailA2A_LsaLD:
    switch ((ncclDevRedOp_t)red) {
    default: WARN("ncclSymkGetKernelIndex: unknown red op %d for id %d", red, (int)id); return -1;
    case ncclDevSum:
      switch (ty) {
      default: WARN("ncclSymkGetKernelIndex: unknown type %d for id %d red %d", (int)ty, (int)id, red); return -1;
      case ncclFloat32: return 16;
      case ncclFloat16: return 27;
      case ncclBfloat16: return 38;
      case ncclFloat8e4m3: return 49;
      case ncclFloat8e5m2: return 60;
      }
    case ncclDevSumPostDiv:
      switch (ty) {
      default: WARN("ncclSymkGetKernelIndex: unknown type %d for id %d red %d", (int)ty, (int)id, red); return -1;
      case ncclFloat32: return 66;
      case ncclFloat16: return 72;
      case ncclBfloat16: return 78;
      case ncclFloat8e4m3: return 84;
      case ncclFloat8e5m2: return 90;
      }
    }
  case ncclSymkKernelId_ReduceScatter_RailA2A_LsaLDMC:
    switch ((ncclDevRedOp_t)red) {
    default: WARN("ncclSymkGetKernelIndex: unknown red op %d for id %d", red, (int)id); return -1;
    case ncclDevSum:
      switch (ty) {
      default: WARN("ncclSymkGetKernelIndex: unknown type %d for id %d red %d", (int)ty, (int)id, red); return -1;
      case ncclFloat32: return 17;
      case ncclFloat16: return 28;
      case ncclBfloat16: return 39;
      case ncclFloat8e4m3: return 50;
      case ncclFloat8e5m2: return 61;
      }
    case ncclDevSumPostDiv:
      switch (ty) {
      default: WARN("ncclSymkGetKernelIndex: unknown type %d for id %d red %d", (int)ty, (int)id, red); return -1;
      case ncclFloat32: return 67;
      case ncclFloat16: return 73;
      case ncclBfloat16: return 79;
      case ncclFloat8e4m3: return 85;
      case ncclFloat8e5m2: return 91;
      }
    }
  }
}
